#define G_LOG_DOMAIN "ap-ability"

#include "ap-ability.h"

#include <math.h>
#include <glib.h>

typedef struct _APAbilityPrivate
{
    gchar *name;
    gchar *label;
    gint   score;
    gint   bonus;
    gint   modifier;
} APAbilityPrivate;

struct _APAbility
{
    GObject parent_instance;
};

G_DEFINE_TYPE_WITH_PRIVATE (APAbility, ap_ability, G_TYPE_OBJECT)

enum
{
    PROP_0,
    PROP_NAME,
    PROP_LABEL,
    PROP_SCORE,
    PROP_BONUS,
    PROP_MODIFER,
    N_PROPERTIES
};

static GParamSpec *obj_properties[N_PROPERTIES] = { NULL, };

void
ap_ability_calculate_modifier (APAbility *self)
{
    APAbilityPrivate *priv = ap_ability_get_instance_private (self);
    gint score;

    /*
     * When a score number will result in a negative number (ex. 9 - 10 = -1)
     * then we subtract by 1 more to return a proper floor ().
     */
    if (priv->score < 10)
        score = priv->score - 1;
    else
        score = priv->score;

    /*
     * Modifer = ((Ability score) - 10) / 2
     */
    priv->modifier = floor ((score - 10) / 2);
}

static void
ap_ability_set_property (GObject      *object,
                         guint         property_id,
                         const GValue *value,
                         GParamSpec   *pspec)
{
    APAbility *self = AP_ABILITY (object);
    APAbilityPrivate *priv = ap_ability_get_instance_private (self);

    switch (property_id)
    {
        case PROP_NAME:
            if (priv->name != NULL)
                g_warning ("Changing the ability name after construct is discouraged. APAbility will not update the name in other widgets.");

            priv->name = g_value_dup_string (value);
            break;

        case PROP_LABEL:
            priv->label = g_value_dup_string (value);
            break;

        case PROP_SCORE:
            priv->score = g_value_get_int (value);
            /* Recalculate modifier */
            ap_ability_calculate_modifier (self);
            break;

        case PROP_BONUS:
            priv->bonus = g_value_get_int (value);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

static void
ap_ability_get_property (GObject      *object,
                         guint         property_id,
                         GValue       *value,
                         GParamSpec   *pspec)
{
    APAbility *self = AP_ABILITY (object);
    APAbilityPrivate *priv = ap_ability_get_instance_private (self);

    switch (property_id)
    {
        case PROP_NAME:
            g_value_set_string (value, priv->name);
            break;

        case PROP_LABEL:
            g_value_set_string (value, priv->label);
            break;

        case PROP_SCORE:
            g_value_set_int (value, priv->score + priv->bonus);
            break;

        case PROP_BONUS:
            g_value_set_int (value, priv->bonus);
            break;

        case PROP_MODIFER:
            g_value_set_int (value, priv->modifier);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

static void
ap_ability_finalize (GObject *object)
{
    G_OBJECT_CLASS (ap_ability_parent_class)->finalize (object);
}

static void
ap_ability_dispose (GObject *object)
{
    G_OBJECT_CLASS (ap_ability_parent_class)->finalize (object);
}

static void
ap_ability_init (APAbility *self)
{
}

static void
ap_ability_class_init (APAbilityClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->set_property = ap_ability_set_property;
    object_class->get_property = ap_ability_get_property;
    object_class->finalize = ap_ability_finalize;
    object_class->dispose = ap_ability_dispose;

    obj_properties[PROP_NAME] =
        g_param_spec_string ("name",
                             "name",
                             "Name of the ability.",
                             NULL,  /* default value */
                             G_PARAM_READWRITE);

    obj_properties[PROP_LABEL] =
        g_param_spec_string ("label",
                             "label",
                             "Display name of the ability.",
                             NULL, /* default value */
                             G_PARAM_READWRITE);

    obj_properties[PROP_SCORE] =
        g_param_spec_int ("score",
                          "score",
                          "The ability score level.",
                          G_MININT,
                          G_MAXINT,
                          0, /* default value */
                          G_PARAM_READWRITE);

    obj_properties[PROP_BONUS] =
        g_param_spec_int ("bonus",
                          "bonus",
                          "The ability score bonus.",
                          G_MININT,
                          G_MAXINT,
                          0, /* default value */
                          G_PARAM_READWRITE);

    obj_properties[PROP_MODIFER] =
        g_param_spec_int ("modifier",
                          "modifier",
                          "The ability modifier based of the ability score.",
                          G_MININT,
                          G_MAXINT,
                          0, /* default value */
                          G_PARAM_READABLE);

    g_object_class_install_properties (object_class,
                                       N_PROPERTIES,
                                       obj_properties);
}

/*
 * *****************************************************************************
 *
 * API
 *
 * *****************************************************************************
 */
APAbility *
ap_ability_new (void)
{
    return g_object_new (AP_TYPE_ABILITY,
                         NULL);
}

void
ap_ability_set_name (APAbility *self,
                     gchar     *name)
{
    GValue new_name = G_VALUE_INIT;

    g_assert (AP_IS_ABILITY (self));
    g_assert (name != NULL);
    g_assert (g_strcmp0 (name, "") != 0);

    g_value_init (&new_name, G_TYPE_STRING);
    g_value_set_string (&new_name, name);

    g_object_set_property (G_OBJECT (self), "name", &new_name);

    g_value_unset (&new_name);
}

gchar *
ap_ability_get_name (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gchar *retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_STRING);

    g_object_get_property (G_OBJECT (self), "name", &val);

    retval = g_value_dup_string (&val);

    g_value_unset (&val);

    return retval;
}

void
ap_ability_set_label (APAbility *self,
                      gchar     *label)
{
    GValue new_label = G_VALUE_INIT;

    g_assert (AP_IS_ABILITY (self));
    g_assert (label != NULL);
    g_assert (g_strcmp0 (label, "") != 0);

    g_value_init (&new_label, G_TYPE_STRING);
    g_value_set_string (&new_label, label);

    g_object_set_property (G_OBJECT (self), "label", &new_label);

    g_value_unset (&new_label);
}

gchar *
ap_ability_get_label (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gchar *retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_STRING);

    g_object_get_property (G_OBJECT (self), "label", &val);

    retval = g_value_dup_string (&val);

    g_value_unset (&val);

    return retval;
}

void
ap_ability_set_score (APAbility *self,
                      gint       score)
{
    GValue new_score = G_VALUE_INIT;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&new_score, G_TYPE_INT);
    g_value_set_int (&new_score, score);

    g_object_set_property (G_OBJECT (self), "score", &new_score);

    g_value_unset (&new_score);
}

gint
ap_ability_get_score (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "score", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}

void
ap_ability_set_bonus (APAbility *self,
                      gint       bonus)
{
    GValue new_bonus = G_VALUE_INIT;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&new_bonus, G_TYPE_INT);
    g_value_set_int (&new_bonus, bonus);

    g_object_set_property (G_OBJECT (self), "bonus", &new_bonus);

    g_value_unset (&new_bonus);
}

gint
ap_ability_get_bonus (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "bonus", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}

gint
ap_ability_get_modifier (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "modifier", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}
